from acai_aws.base.event import BaseRecordsEvent
from acai_aws.msk.record import Record


class Event(BaseRecordsEvent):

    def __init__(self, event, context=None, **kwargs):
        super().__init__(event, context, **kwargs)
        self._record_class = Record

    @property
    def raw_records(self):
        return self._event.get('records', {})

    @property
    def records(self):
        self._records = []
        for topic in self.raw_records:
            for msk_record in self.raw_records[topic]:
                self._records.append(self._record_class(msk_record))
        self._validate_operations()
        self._validate_record_body()
        return self.data_classes if self.data_class is not None else self._records

    @property
    def topics(self):
        return list(self.raw_records.keys())



from collections import defaultdict

from jsonschema import Draft7Validator

from acai_aws.common.schema import Schema


class Validator:

    def __init__(self, **kwargs):
        self.__schema = Schema(**kwargs)
        self.__pairings = {
            'required_headers': 'headers',
            'available_headers': 'headers',
            'required_query': 'query_params',
            'available_query': 'query_params',
            'required_body': 'body'
        }

    def auto_load(self):
        self.__schema.load_schema_file()

    def request_has_security(self, request):
        route_spec = self.__schema.get_route_spec(request.route, request.method)
        if route_spec.get('security'):
            return True
        return False

    def validate_request_with_openapi(self, request, response):
        route_spec = self.__schema.get_route_spec(request.route, request.method)
        requirements = Validator.combine_parameters(route_spec.get('parameters', []))
        if route_spec.get('requestBody'):
            requirements['required_body'] = route_spec['requestBody']['content'][request.content_type]['schema']
        self.validate_request(request, response, requirements)

    def validate_request(self, request, response, requirements):
        for required, source in self.__pairings.items():
            if requirements.get(required) and required == 'required_body':
                Validator.check_required_body(response, self.__schema.get_body_spec(requirements[required]), getattr(request, source))
            elif requirements.get(required) and 'required' in required:
                Validator.check_required_fields(response, requirements[required], getattr(request, source), source)
            elif requirements.get(required) and 'available' in required:
                full_list = Validator.combine_available_with_required(requirements, required)
                Validator.check_available_fields(response, full_list, getattr(request, source), source)
        if response.has_errors:
            response.code = 400

    def validate_response_with_openapi(self, request, response):
        requirements = {}
        route_spec = self.__schema.get_route_spec(request.route, request.method)
        if route_spec.get('responses'):
            requirements['required_response'] = route_spec['responses'][f'{response.code}']['content'][response.content_type]['schema']
        self.validate_response(response, requirements)

    def validate_response(self, response, requirements):
        Validator.check_required_body(response, self.__schema.get_body_spec(requirements.get('required_response', {})), response.raw)
        if response.has_errors:
            response.set_error('response', 'There was a problem with the APIs response; does not match defined schema')
            response.code = 500

    def validate_record_body(self, body, schema):
        errors = []
        schema_validator = Draft7Validator(self.__schema.get_body_spec(schema))
        for schema_error in sorted(schema_validator.iter_errors(body), key=str):
            error_key = Validator.format_schema_error_key(schema_error)
            errors.append({'key': error_key, 'message': schema_error.message})
        return errors

    @staticmethod
    def check_required_fields(response, required, sent, list_name=''):
        sent_keys = []
        if isinstance(sent, dict) and len(sent.keys()) > 0:
            sent_keys = sent.keys()
        missing_fields = [value for value in required if value not in sent_keys]
        if len(required) > 0:
            for field in missing_fields:
                response.set_error(list_name, f'Please provide {field} in {list_name}')

    @staticmethod
    def check_available_fields(response, available, sent, list_name=''):
        if len(available) > 0:
            unavailable_fields = [value for value in sent if value not in available]
            for field in unavailable_fields:
                response.set_error(list_name, f'{field} is not an available {list_name}')

    @staticmethod
    def combine_available_with_required(requirements, required):
        avail_list = requirements[required]
        if required == 'available_query' and requirements.get('required_query'):
            avail_list += requirements['required_query']
        elif required == 'available_headers' and requirements.get('required_headers'):
            avail_list += requirements['required_headers']
        return avail_list

    @staticmethod
    def check_required_body(response, schema, request_body):
        if schema:
            schema_validator = Draft7Validator(schema)
            for schema_error in sorted(schema_validator.iter_errors(request_body), key=str):
                error_key = Validator.format_schema_error_key(schema_error)
                response.set_error(key_path=error_key, message=schema_error.message)

    @staticmethod
    def combine_parameters(parameters):
        requirements = defaultdict(lambda: [])
        for param in parameters:
            if param.get('in') == 'query' and param.get('required'):
                requirements['required_query'].append(param['name'])
            elif param.get('in') == 'query':
                requirements['available_query'].append(param['name'])
            elif param.get('in') == 'header' and param.get('required'):
                requirements['required_headers'].append(param['name'])
            elif param.get('in') == 'header':
                requirements['available_headers'].append(param['name'])
        return dict(requirements)

    @staticmethod
    def format_schema_error_key(schema_error):
        error_path = '.'.join(str(path) for path in schema_error.path)
        return error_path if error_path else 'root'



import unittest

from acai_aws.common.records.exception import RecordException

from tests.mocks.dynamodb import mock_event
from tests.mocks.dynamodb.mock_functions import mock_ddb_full, mock_ddb_basic, mock_ddb_operation_ignore, mock_ddb_operation_raise, before_call, after_call, \
    call_list


class DynamoDBRequirementsTest(unittest.TestCase):
    created_record = mock_event.get_created_event()

    def test_ddb_decorator_with_basic_requirements(self):
        expected = {'ddb_basic': ['INSERT']}
        result = mock_ddb_basic(self.created_record, None)
        self.assertDictEqual(result, expected)

    def test_ddb_decorator_with_full_requirements(self):
        expected = {'ddb_full': [True]}
        result = mock_ddb_full(self.created_record, None)
        self.assertDictEqual(result, expected)
        self.assertTrue(before_call.has_been_called)
        self.assertTrue(after_call.has_been_called)
        self.assertEqual('before', call_list[0])
        self.assertEqual('after', call_list[1])

    def test_ddb_decorator_ignore_non_matching_operation(self):
        expected = {'ddb_operation_ignore': []}
        result = mock_ddb_operation_ignore(self.created_record, None)
        self.assertDictEqual(result, expected)

    def test_ddb_decorator_ignore_non_matching_operation_error(self):
        try:
            mock_ddb_operation_raise(self.created_record, None)
            self.assertTrue(False)
        except RecordException as record_error:
            self.assertTrue(isinstance(record_error, RecordException))



from acai_aws.apigateway.exception import ApiException


class ConfigValidator:

    @staticmethod
    def validate(**kwargs):
        if not kwargs.get('base_path') or not isinstance(kwargs.get('base_path'), str):
            raise ApiException(code=500, message='base_path string is required')
        if not kwargs.get('handlers') or not isinstance(kwargs.get('handlers'), (str, dict)):
            raise ApiException(code=500, message='handlers is required; must be glob pattern string or dictionary mapping {route: file_path}')
        if kwargs.get('schema') and not isinstance(kwargs.get('schema'), (str, dict)):
            raise ApiException(code=500, message='schema should either be file path string or json-schema style dictionary')
        if kwargs.get('auto_validate') and not isinstance(kwargs.get('auto_validate'), bool):
            raise ApiException(code=500, message='auto_validate should be a boolean')
        if kwargs.get('validate_response') and not isinstance(kwargs.get('validate_response'), bool):
            raise ApiException(code=500, message='validate_response should be a boolean')
        if kwargs.get('cache_size') and not isinstance(kwargs.get('cache_size'), int) and kwargs.get('cache_size') is not None:
            raise ApiException(code=500, message='cache_size should be an int (0 for unlimited size) or None (to disable route caching)')
        if kwargs.get('cache_mode') and kwargs['cache_mode'] not in ('all', 'static-only', 'dynamic-only'):
            raise ApiException(code=500, message='cache_mode should be a string of the one of the following values: all, static-only, dynamic-only')
        if kwargs.get('verbose_logging') and not isinstance(kwargs.get('verbose_logging'), bool):
            raise ApiException(code=500, message='verbose_logging should be a boolean')



from acai_aws.common.records.exception import RecordException
from acai_aws.base.no_data import NoDataClass
from acai_aws.base.placeholder import PlaceHolderRecord
from acai_aws.common.validator import Validator


class BaseRecordsEvent:

    def __init__(self, event, context=None, **kwargs):
        self._event = event
        self._context = context
        self._kwargs = kwargs
        self._records = []
        self._record_class = PlaceHolderRecord
        self.__data_class = NoDataClass
        self.__validator = Validator(**kwargs)

    @property
    def event(self):
        return self._event

    @property
    def context(self):
        return self._context

    @property
    def raw_records(self):
        return self._event.get('Records', [])

    @property
    def data_class(self):
        if issubclass(self.__data_class, NoDataClass):
            return None
        return self.__data_class

    @data_class.setter
    def data_class(self, data_class):
        self.__data_class = data_class

    @property
    def data_classes(self):
        return [self.data_class(record=record) for record in self._records]

    @property
    def records(self):
        self._records = [self._record_class(record) for record in self.raw_records]
        self._validate_operations()
        self._validate_record_body()
        return self.data_classes if self.data_class is not None else self._records

    def _validate_operations(self):
        if not self._kwargs.get('operations'):
            return
        validated = []
        for record in self._records:
            if record.operation in self._kwargs['operations']:
                validated.append(record)
            elif self._kwargs.get('raise_operation_error'):
                raise RecordException(record=record, message=f'record did not meet operation requirement; required: {self._kwargs["operations"]}, received: {record.operation}')
        self._reset_records(validated)

    def _validate_record_body(self):
        if not self._kwargs.get('required_body'):
            return
        validated = []
        for record in self._records:
            errors = self.__validator.validate_record_body(record.body, self._kwargs.get('required_body'))
            if len(errors) != 0 and self._kwargs.get('raise_body_error'):
                raise RecordException(record=record, message=f'record did not meet body requirement; errors: {errors}')
            if len(errors) == 0:
                validated.append(record)
        self._reset_records(validated)

    def _reset_records(self, validated):
        self._records.clear()
        self._records = validated

    def __str__(self):
        return str([str(record) for record in self.records])



from acai_aws.base.event import BaseRecordsEvent
from acai_aws.sqs.record import Record


class Event(BaseRecordsEvent):

    def __init__(self, event, context=None, **kwargs):
        super().__init__(event, context, **kwargs)
        self._record_class = Record



import unittest

from acai_aws.apigateway.request import Request
from acai_aws.apigateway.response import Response
from acai_aws.apigateway.resolver.modes.pattern import PatternModeResolver
from acai_aws.apigateway.exception import ApiException
from tests.mocks.apigateway import mock_request


class PatternModeResolverTest(unittest.TestCase):
    basic_request = mock_request.get_basic()
    nested_request = mock_request.get_basic_nested()
    init_request = mock_request.get_basic_init()
    dynamic_request = mock_request.get_dynamic()
    bad_route_request = mock_request.get_bad_route()
    triple_request = mock_request.get_triple_post()
    base_path = 'unit-test/v1'
    pattern = 'tests/mocks/apigateway/resolver/pattern_handlers/**/*_controller.py'
    expected_endpoint_return = {
        'hasErrors': False,
        'response': {
            'headers': {
                'Access-Control-Allow-Origin': '*',
                'Access-Control-Allow-Headers': '*'
            },
            'statusCode': 200,
            'isBase64Encoded': False,
            'body': {
                'basic_pattern': True
            }
        }
    }

    def setUp(self):
        self.pattern_resolver = PatternModeResolver(base_path=self.base_path, handlers=self.pattern)

    def test_get_endpoint_module(self):
        request = Request(self.basic_request)
        response = Response()
        endpoint_module = self.pattern_resolver.get_endpoint_module(request)
        self.assertTrue(hasattr(endpoint_module, 'post'))
        endpoint_returns = endpoint_module.post(request, response)
        self.assertEqual(str(self.expected_endpoint_return), str(endpoint_returns))

    def test_basic_get_file_and_import_path(self):
        request = Request(self.basic_request)
        file_path, import_path = self.pattern_resolver._get_file_and_import_path(request.path)
        self.assertTrue('tests/mocks/apigateway/resolver/pattern_handlers/basic/basic_controller.py' in file_path)
        self.assertEqual('tests.mocks.apigateway.resolver.pattern_handlers.basic.basic_controller', import_path)

    def test_nested_get_file_and_import_path(self):
        request = Request(self.nested_request)
        file_path, import_path = self.pattern_resolver._get_file_and_import_path(request.path)
        self.assertTrue('tests/mocks/apigateway/resolver/pattern_handlers/nested_1/nested_2/basic/basic_controller.py' in file_path)
        self.assertEqual('tests.mocks.apigateway.resolver.pattern_handlers.nested_1.nested_2.basic.basic_controller', import_path)

    def test_default_init_get_file_and_import_path(self):
        request = Request(self.init_request)
        file_path, import_path = self.pattern_resolver._get_file_and_import_path(request.path)
        self.assertTrue('tests/mocks/apigateway/resolver/pattern_handlers/home/home_controller.py' in file_path)
        self.assertEqual('tests.mocks.apigateway.resolver.pattern_handlers.home.home_controller', import_path)

    def test_dynamic_get_file_and_import_path(self):
        request = Request(self.dynamic_request)
        file_path, import_path = self.pattern_resolver._get_file_and_import_path(request.path)
        self.assertTrue('tests/mocks/apigateway/resolver/pattern_handlers/dynamic/_id_controller.py' in file_path)
        self.assertEqual('tests.mocks.apigateway.resolver.pattern_handlers.dynamic._id_controller', import_path)

    def test_file_and_import_path_not_found_raises_resolver_exception(self):
        try:
            request = Request(self.bad_route_request)
            self.pattern_resolver._get_file_and_import_path(request.path)
            self.assertTrue(False)
        except ApiException as resolver_error:
            self.assertTrue(isinstance(resolver_error, ApiException))
            self.assertEqual(resolver_error.code, 404)
            self.assertEqual(resolver_error.message, 'route not found')

    def test_triple_dynamic_get_file_and_import_path(self):
        request = Request(self.triple_request)
        file_path, import_path = self.pattern_resolver._get_file_and_import_path(request.path)
        self.assertTrue('tests/mocks/apigateway/resolver/pattern_handlers/triple/_coordinates_controller.py' in file_path)
        self.assertEqual('tests.mocks.apigateway.resolver.pattern_handlers.triple._coordinates_controller', import_path)

    def test_single_nested_dynamic_get_file_and_import_path(self):
        dynamic_nested_request = mock_request.get_dynamic_nested_request_get('user/1')
        request = Request(dynamic_nested_request)
        file_path, import_path = self.pattern_resolver._get_file_and_import_path(request.path)
        self.assertTrue('tests/mocks/apigateway/resolver/pattern_handlers/user/_user_id/_user_id_controller.py' in file_path)
        self.assertEqual('tests.mocks.apigateway.resolver.pattern_handlers.user._user_id._user_id_controller', import_path)

    def test_double_nested_dynamic_get_file_and_import_path(self):
        dynamic_nested_request = mock_request.get_dynamic_nested_request_get('user/1/item')
        request = Request(dynamic_nested_request)
        file_path, import_path = self.pattern_resolver._get_file_and_import_path(request.path)
        self.assertTrue('tests/mocks/apigateway/resolver/pattern_handlers/user/_user_id/item/item_controller.py' in file_path)
        self.assertEqual('tests.mocks.apigateway.resolver.pattern_handlers.user._user_id.item.item_controller', import_path)

    def test_triple_nested_dynamic_get_file_and_import_path(self):
        dynamic_nested_request = mock_request.get_dynamic_nested_request_get('user/1/item/a')
        request = Request(dynamic_nested_request)
        file_path, import_path = self.pattern_resolver._get_file_and_import_path(request.path)
        self.assertTrue('tests/mocks/apigateway/resolver/pattern_handlers/user/_user_id/item/_item_id_controller.py' in file_path)
        self.assertEqual('tests.mocks.apigateway.resolver.pattern_handlers.user._user_id.item._item_id_controller', import_path)



import unittest

from acai_aws.apigateway.config_validator import ConfigValidator
from acai_aws.apigateway.exception import ApiException


class ConfigValidatorTest(unittest.TestCase):

    def test_config_validator_validates_all_passing(self):
        try:
            ConfigValidator.validate(
                base_path='some/path',
                handlers={'route': 'file/path.py'},
                schema='path/to/schema',
                auto_validate=True,
                validate_response=True,
                cache_size=128,
                cache_mode='all'
            )
            # alternative options
            ConfigValidator.validate(
                base_path='some/path',
                handlers='some/path',
                schema={'json_schema': True},
                auto_validate=False,
                validate_response=False,
                cache_size=None
            )
            self.assertTrue(True)
        except ApiException as api_error:
            print(api_error.message)
            self.assertTrue(False)

    def test_config_validator_validates_base_path(self):
        try:
            ConfigValidator.validate(**{})
        except ApiException as api_error:
            self.assertTrue(isinstance(api_error, ApiException))
            self.assertEqual('base_path string is required', api_error.message)

    def test_config_validator_validates_routing_handlers_is_required(self):
        try:
            ConfigValidator.validate(base_path='some/path')
            self.assertTrue(False)
        except ApiException as api_error:
            self.assertTrue(isinstance(api_error, ApiException))
            self.assertEqual('handlers is required; must be glob pattern string or dictionary mapping {route: file_path}', api_error.message)

    def test_config_validator_validates_routing_handlers_are_appropriate(self):
        try:
            ConfigValidator.validate(base_path='some/path', handlers=1)
            self.assertTrue(False)
        except ApiException as api_error:
            self.assertTrue(isinstance(api_error, ApiException))
            self.assertEqual('handlers is required; must be glob pattern string or dictionary mapping {route: file_path}', api_error.message)

    def test_config_validator_validates_routing_schema_is_appropriate(self):
        try:
            ConfigValidator.validate(base_path='some/path', handlers={'route': 'file/path.py'}, schema=1)
            self.assertTrue(False)
        except ApiException as api_error:
            self.assertTrue(isinstance(api_error, ApiException))
            self.assertEqual('schema should either be file path string or json-schema style dictionary', api_error.message)

    def test_config_validator_validates_routing_auto_validate_is_appropriate(self):
        try:
            ConfigValidator.validate(base_path='some/path', handlers={'route': 'file/path.py'}, auto_validate=1)
            self.assertTrue(False)
        except ApiException as api_error:
            self.assertTrue(isinstance(api_error, ApiException))
            self.assertEqual('auto_validate should be a boolean', api_error.message)

    def test_config_validator_validates_routing_validate_response_is_appropriate(self):
        try:
            ConfigValidator.validate(base_path='some/path', handlers={'route': 'file/path.py'}, validate_response=1)
            self.assertTrue(False)
        except ApiException as api_error:
            self.assertTrue(isinstance(api_error, ApiException))
            self.assertEqual('validate_response should be a boolean', api_error.message)

    def test_config_validator_validates_routing_verbose_logging_is_appropriate(self):
        try:
            ConfigValidator.validate(base_path='some/path', handlers={'route': 'file/path.py'}, verbose_logging=1)
            self.assertTrue(False)
        except ApiException as api_error:
            self.assertTrue(isinstance(api_error, ApiException))
            self.assertEqual('verbose_logging should be a boolean', api_error.message)

    def test_config_validator_validates_routing_cache_size_is_appropriate(self):
        try:
            ConfigValidator.validate(base_path='some/path', handlers={'route': 'file/path.py'}, cache_size='1')
            self.assertTrue(False)
        except ApiException as api_error:
            self.assertTrue(isinstance(api_error, ApiException))
            self.assertEqual('cache_size should be an int (0 for unlimited size) or None (to disable route caching)', api_error.message)

    def test_config_validator_validates_routing_cache_mode_is_appropriate(self):
        try:
            ConfigValidator.validate(base_path='some/path', handlers={'route': 'file/path.py'}, cache_mode='bad')
            self.assertTrue(False)
        except ApiException as api_error:
            self.assertTrue(isinstance(api_error, ApiException))
            self.assertEqual('cache_mode should be a string of the one of the following values: all, static-only, dynamic-only', api_error.message)



class Endpoint:

    def __init__(self, module, method):
        self.__method = getattr(module, method)
        self.__requirements = getattr(self.__method, 'requirements', {})

    @property
    def has_requirements(self):
        return bool(self.__requirements)

    @property
    def requirements(self):
        return self.__requirements

    @property
    def requires_auth(self):
        return self.__requirements.get('auth_required', False)

    @property
    def has_required_response(self):
        return self.__requirements.get('required_response', False)

    @property
    def has_required_route(self):
        return bool(self.__requirements.get('required_route'))

    @property
    def required_route(self):
        return self.__requirements.get('required_route', '')

    def run(self, request, response):
        return self.__method(request, response)



from acai_aws.common.json_helper import JsonHelper
from acai_aws.base.record import BaseRecord


class Record(BaseRecord):

    @property
    def version(self):
        return self._record.get('EventVersion')

    @property
    def subscription_arn(self):
        return self._record.get('EventSubscriptionArn')

    @property
    def source(self):
        return self._record.get('EventSource')

    @property
    def signature_version(self):
        return self._record.get('Sns', {}).get('SignatureVersion')

    @property
    def timestamp(self):
        return self._record.get('Sns', {}).get('Timestamp')

    @property
    def signature(self):
        return self._record.get('Sns', {}).get('Signature')

    @property
    def signing_cert_url(self):
        return self._record.get('Sns', {}).get('SigningCertUrl')

    @property
    def message_id(self):
        return self._record.get('Sns', {}).get('MessageId')

    @property
    def message(self):
        return self._record.get('Sns', {}).get('Message')

    @property
    def message_attributes(self):
        return self._record.get('Sns', {}).get('MessageAttributes')

    @property
    def attributes(self):
        if not self._attributes:
            for key in self.message_attributes:
                self._attributes[key] = self.message_attributes[key].get('Value')
        return self._attributes

    @property
    def sns_type(self):
        return self._record.get('Sns', {}).get('Type')

    @property
    def unsubscribe_url(self):
        return self._record.get('Sns', {}).get('UnsubscribeUrl')

    @property
    def topic_arn(self):
        return self._record.get('Sns', {}).get('TopicArn')

    @property
    def subject(self):
        return self._record.get('Sns', {}).get('Subject')

    @property
    def body(self):
        return JsonHelper.decode(self.message)

    @property
    def operation(self):
        return self.UNKNOWN

    def __str__(self):
        return str({
            'version': self.version,
            'subscription_arn': self.subscription_arn,
            'source': self.source,
            'timestamp': self.timestamp,
            'signature': self.signature,
            'signing_cert_url': self.signing_cert_url,
            'message_id': self.message_id,
            'message_attributes': self.message_attributes,
            'message': self.message,
            'attributes': self.attributes,
            'sns_type': self.sns_type,
            'unsubscribe_url': self.unsubscribe_url,
            'topic_arn': self.topic_arn,
            'body': self.body,
            'operation': self.operation
        })
